
(* prints error message on stderr and aborts *)
fun emit_error m = let open ErrPrint in print "emit error: "; print m; print "\n"; flush (); raise BYE end;

(* toplevel emit routine *)
signature EMIT = sig
val emit : TTS -> unit
end;


	  

(*
applied before emit, normalizes the types of state variables,
introduicing the necessary type declarations;
- interval and nat types replaced by int;
- unions, records and queues are named (but not arrays);
*)

signature NAME = sig
  val nameTypes : TTS -> TTS
end

structure Name :> NAME = struct

(* utilities *)

(* lexicographic list comparisons *)
fun rcompare _ [] (_ :: _) = LESS
  | rcompare _ (_ :: _) [] = GREATER
  | rcompare cmp (h1 :: t1) (h2 :: t2) = (case cmp (h1,h2) of EQUAL => rcompare cmp t1 t2 | b => b)
  | rcompare _ [] [] = EQUAL;

(* used to creates new names for types and channels *)
val newName =
    let val no = ref 0
    in fn prefix =>
           let val n = prefix ^ Int.toString (!no)
           in no := 1 + !no;
              n
           end
    end;

(* holds processed declarations *)
val DECLS = ref ([] : (string * Type) list);


(* type keyed avl of strings *)
structure typeKey : KEY = struct
type key = Type;
fun level ty =  (* identifies all numeric types *)
 case ty of
	  BoolType => 1
	| IntType => 2
	| CharType => 2
	| UnsignedCharType => 2
	| ShortType => 2
	| UnsignedShortType => 2
	| NatType => 2
	| NamedType _ => 3
	| EnumType _ => 4
	| UnionType _ => 5
	| RecordType _ => 6
	| ArrayType _ => 7
	| QueueType _ => 8
        | _ => (* unexpected *) 20;

fun compare (x,y) = case Int.compare (level x,level y) of
	  EQUAL => (case (x,y) of
		(NamedType s1, NamedType s2) => String.compare (s1,s2) 
	      | (EnumType cl1, EnumType cl2) =>	rcompare (fn (c1,c2) => String.compare (c1,c2)) cl1 cl2
	      | (UnionType ctyl1, UnionType ctyl2) =>
		rcompare (fn ((c1,ty1),(c2,ty2)) =>
			     case String.compare (c1,c2) of
				 EQUAL => (case (ty1,ty2) of
					       (NONE,NONE) => EQUAL
					     | (NONE,_) => LESS
					     | (_,NONE) => GREATER
					     | (SOME ty1,SOME ty2) => compare(ty1,ty2))
			       | b => b)
			 ctyl1
			 ctyl2
	      | (RecordType ftyl1, RecordType ftyl2) =>
		rcompare (fn ((f1,ty1),(f2,ty2)) => case String.compare (f1,f2) of EQUAL => compare(ty1,ty2) | b => b) ftyl1 ftyl2
	      | (ArrayType (sz1,ty1), ArrayType (sz2,ty2)) =>
		(case Int.compare (sz1,sz2) of EQUAL => compare (ty1,ty2) | b => b)
	      | (QueueType (sz1,ty1), QueueType (sz2,ty2)) =>
		(case Int.compare (sz1,sz2) of EQUAL => compare (ty1,ty2) | b => b)
	      | _ => EQUAL)
	| v => v;

fun x < y = case compare (x,y) of LESS => true | _ => false;
fun x > y = case compare (x,y) of GREATER => true | _ => false;
end;
structure tyAvlTree = avlTreef (typeKey);

(* names avl : type -> name, if any *)
val TYNAMES = tyAvlTree.newTree() : String.string tyAvlTree.tree ref;
fun insertType ty n = TYNAMES := tyAvlTree.insert (ty,n,!TYNAMES);
fun lookupType ty = tyAvlTree.getData (ty,!TYNAMES);

(* type ty was found equivalent to n (modulo arithmetic types), updates definition of n if not equal *)
fun adjustTypeDecl n ty =
    let infix lub
	fun ty1 lub ty2 = (* ty2 can be named *)
	    case (ty1,ty2) of
	      (* one is signed *)
		(IntType,_) => IntType
	      | (_,IntType) => IntType
	      | (ShortType,_) => ShortType
	      | (_,ShortType) => ShortType
	      | (CharType,_) => CharType
	      | (_,CharType) => CharType
	      (* none is signed *)
	      | (NatType,_) => NatType
	      | (_,NatType) => NatType
	      | (UnsignedShortType,_) => UnsignedShortType
	      | (_,UnsignedShortType) => UnsignedShortType
	      | (UnsignedCharType,_) => UnsignedCharType
	      | (_,UnsignedCharType) => UnsignedCharType
	      (* structures *)
	      | (RecordType flty1, RecordType flty2) => RecordType (map (fn ((f,ty1),(_,ty2)) => (f,ty1 lub ty2)) (ListPair.zip (flty1,flty2)))
	      | (ArrayType (sz,ty1), ArrayType (_,ty2)) => ArrayType(sz,ty1 lub ty2)
	      | (QueueType (sz,ty1), QueueType (_,ty2)) => QueueType(sz,ty1 lub ty2)
	      | _ => ty2
	val ty' = assoc n (!DECLS)
	fun upd [] = [] | upd ((nty as (n',_))::t) = if n=n' then (n',ty lub ty')::t else nty::upd t
    in DECLS := upd (!DECLS)
    end;
    
(* creates a name for a type and its constituents, making sure names are new *)
fun nameType pref ty =
    (* if ty is named, return its name, otherwise generate a new name and stores it *)
    case lookupType ty of 
	SOME n => (adjustTypeDecl n ty; NamedType n)
      | _ => let val n = newName pref
	     in DECLS := (n,ty) :: !DECLS;
		insertType ty n;
		NamedType n
	     end;

(* returns alias for ty, unless is a constant type *)
(* constants have been computed, and names expanded *)
(* first version specializes representations of intervals (to save space) and keeps NatType.
   This interferes with subtyping.
   optimal solution would be to add casts in the code generated when a value of a named type is promoted to its supertype,
   suboptimal solution here is to keep only the lubs of named types in the < relation (nameType handles that) *)
fun aliasType ty =
    case ty of
        IntervalType (a,b) =>
	((* if a >= 0
	 then if b < 256 then UnsignedCharType else
	      if b < 65536 then UnsignedShortType else
	      NatType
	 else *) if a >= ~127 andalso b < 128 then CharType else
	      if a >= ~32768 andalso b < 32768 then ShortType else
	      IntType)
      | NatType => IntType
      | EnumType cl => nameType "e" ty
      | UnionType ctyl => nameType "u" (UnionType (map (fn (f,SOME ty) => (f,SOME (aliasType ty)) | co => co) ctyl))
      | RecordType ftyl => nameType "r" (RecordType (map (fn (f,ty) => (f,aliasType ty)) ftyl))
      | ArrayType (sz,ety) => nameType "a" (ArrayType (sz,aliasType ety))
      | QueueType (sz,ety) => nameType "q" (QueueType (sz,aliasType ety))
      | _ => ty;

(* names types in embedded StBlocks *)
fun nameInExp e = (* note: only types occurring in declarations are named (except in special cases ...) *)
    case e of
 	InfixExp(EQ,ty,e1,e2) => InfixExp(EQ,aliasType ty,nameInExp e1,nameInExp e2)
      | InfixExp(NE,ty,e1,e2) => InfixExp(NE,aliasType ty,nameInExp e1,nameInExp e2)
      | InfixExp(f,ty,e1,e2) => InfixExp(f,ty,nameInExp e1,nameInExp e2)
      | BinopExp(f,ty,e1,e2) => BinopExp(f,ty,nameInExp e1,nameInExp e2)
      | UnopExp(f,ty,e) => UnopExp(f,ty,nameInExp e)
      | RecordExp(fl,ty) => RecordExp(map (fn (f,e) => (f,nameInExp e)) fl,ty)
      | ArrayExp(el,ty) => ArrayExp(map nameInExp el,ty)
      | QueueExp(el,ty) => QueueExp(map nameInExp el,ty)
      | ArrayAccessExp(e1,e2,ty) => ArrayAccessExp(nameInExp e1,nameInExp e2,ty)
      | RecordAccessExp(e,f,ty) => RecordAccessExp(nameInExp e,f,ty)
      | CondExp(c,e1,e2) => CondExp(nameInExp c,nameInExp e1,nameInExp e2)
      | ConstrExp(c,e,ty) => ConstrExp(c,nameInExp e,ty)
      | AppExp(c,el,tyl,ty) => AppExp(c,map nameInExp el,tyl,aliasType ty)  (* aliased type needed for CALLFUN *)
      | CallExp(c,el,tyl,ty) => CallExp(c,map nameInExp el,tyl,aliasType ty)  (* aliased type needed for CALLFUN *)
      | DestrExp (c,e) => DestrExp (c,nameInExp e)
      | MatchesExp(c,e) => MatchesExp(c,nameInExp e)
      | LetExp(s,e) => LetExp(nameInSt s,nameInExp e)
      | _ => e
and nameInSt s =
    case s of
	StGuard e => StGuard (nameInExp e)
      | StBlock(dl,s) => StBlock (map (fn (s,ty) => (s,aliasType ty)) dl, nameInSt s)
      | StSequence(s1,s2) => StSequence(nameInSt s1, nameInSt s2)
      | StWhile(e,s) => StWhile(nameInExp e, nameInSt s)
      | StForeach(v,s) => StForeach(v, nameInSt s)  (* note: type NOT named in v *)
      | StCond(e,s1,s2) => StCond(nameInExp e, nameInSt s1, nameInSt s2)
      | StAssign(p,e) => StAssign(nameInExp p,nameInExp e)
      | StReturn e => StReturn(nameInExp e)
      | _ => s;

(* names types in declarations and in constants (so that enum/union declarations are not lost!) *)
(* added type declarations prefix first section *)
fun nameTypes (Program sl) =
    let fun nameDecl d =
	    case d of
		ProcessDecl (b,c,d,e,rkl,kl,vl,f,g,init,h) =>
		let val rkl' = map (fn (s,ty,e) => (s,aliasType ty,nameInExp e)) rkl
		    val kl' = map (fn (s,ty,e) => (s,aliasType ty,nameInExp e)) kl
		    val vl' = map (fn (s,ty,e) => (s,aliasType ty,nameInExp e)) vl
		    val h' = List.map (fn (tg,fr,s,b,p) => (tg,fr,nameInSt s,b,p)) h
		in ProcessDecl (b,c,d,e,rkl',kl',vl',f,g,init,h')
		end
	      | FunctionDecl (f,args,ty,vars,body) =>
		FunctionDecl (f,
			      map (fn (v,ty) => (v,aliasType ty)) args,
			      aliasType ty,
			      map (fn (s,ty,e) => (s,aliasType ty,nameInExp e)) vars,
			      nameInSt body)
	      | ExternDecl (f,argtypes,ty) => ExternDecl (f,map aliasType argtypes,aliasType ty)
	      | EventDecl (t,sym,tyl) => EventDecl (t,sym,map aliasType tyl)
	      | TaskDecl (t,sym,argtypes,ty) => TaskDecl (t,sym,map aliasType argtypes,aliasType ty)					      
	      | d => d
	fun nameSection (s,dl,m) =
	    (s, map nameDecl dl, nameDecl m)
	val ((s1,dl1,m1)::sect) = map nameSection sl
    in List.app (fn (_,_,_,r) => r := nameInSt (!r)) (!VALUEVARS);
       if HPX andalso !HIPPOEXT then 
	   (* hippo only *)
	   TYCH_DECLS := List.map (fn TypeDecl (n,ty) => TypeDecl (n,aliasType ty)
				    | ChannelDecl (n,tyl) => ChannelDecl (n,List.map aliasType tyl)
				    | d => d)
				  (List.rev (!TYCH_DECLS))
       else
	   ();			      
       Program ((s1,map TypeDecl (rev (!DECLS)) @ dl1,m1)::sect)
    end;

end; (* structure Name *)




